@@ -30,7 +30,6 @@ enum file_format {
 	FASTA = 1,
 	FASTA_CONT,
 	FASTQ,
-	INTERLEAVED,
 	BAM,
 	TAB_MATE5,
 	TAB_MATE6,
@@ -44,7 +43,6 @@ static const std::string file_format_names[] = {
 	"FASTA",
 	"FASTA sampling",
 	"FASTQ",
-	"Interleaved FASTQ",
 	"BAM",
 	"Tabbed mated",
 	"Raw",
